//////////////////////////////////////////////////////////////////////
//                                                                  //
//  Clueless - an noai-AI                                           //
//                                                                  //
//////////////////////////////////////////////////////////////////////
// 
// Clueless need some help from you in order to function. First, and
// most important he would be happy if you turn off build-on-slope in 
// the patch-settings. This is because he can only build roads from 
// center to center (of tiles).
// 
// Also he prefers flat, smooth maps with low sea-level. The map should
// have low amount of towns and no industries. Industries are big and 
// scary. Clueless don't know how to build around them.
//
// He haven't interacted that much with players so be kind to him.  
// Don't build nasty railways that block his way. He don't know how
// to cross them yet.
//
//
// Author: Zuu (Leif Linse), www.zuu.info, user Zuu @ tt-forums.net
// Purpose: To play around with the noai frame work.
//               - Not to make the next big thing!
// License: GNU GPL



//////////////////////////////////////////////////////////////////////
//                                                                  //
//  CLASS: Helper                                                   //
//                                                                  //
//////////////////////////////////////////////////////////////////////
class Helper
{
	base     = null;
	cargo    = null;
	company  = null;
	industry = null;
	map      = null;
	sign     = null;
	town     = null;

	company_name = null; // a copy of state_company_name which does not include the state-prefix that is included if one ask company.GetCompanyName();

	constructor() {
		this.base = AIBase();
		this.cargo = AICargo();
		this.company = AICompany();
		this.industry = AIIndustry();
		this.map = AIMap();
		this.sign = AISign();
		this.town = AITown();

		this.company_name = "";
	}

	function Distance(a, b);
	function TownDistance(town1, town2);
	function StepFunction(t);
	function ArrayFind(array, toFind);
	function TileLocationString(tile);
	function ClearAllSigns();
	function IsTownInConnectionList(connection_list, check_town);
	function Debug(text);
}

function Helper::TownDistance(town1, town2)
{
	return this.map.DistanceManhattan(this.town.GetLocation(town1), this.town.GetLocation(town2));
}
function Helper::Distance(a, b)
{
	return this.map.DistanceManhattan(a, b);
}
function Helper::StepFunction(t)
{
	if(t>=0)
		return 1;

	return 0;
}
function Helper::TileLocationString(tile)
{
	return "(" + this.map.GetTileX(tile) + ", " + this.map.GetTileY(tile) + ")";
}
function Helper::ClearAllSigns()
{
	local i = 0;
	for(i = 0; i < this.sign.GetMaxSignID(); ++i)
	{
		if(i != null && this.sign.IsValidSign(i))
		{
			this.sign.RemoveSign(i);
		}
	}
}
function Helper::IsTownInConnectionList(connection_list, check_town)
{
	foreach(val in connection_list)
	{
		foreach(town in val.town)
		{
			if(town == check_town)
			{
				return true;
			}
		}
	}
	return false;
}
function Helper::Debug(text)
{
	//print(this.company_name + ": " + text);
}

// RETURN null if not found, else the key to the found value.
function Helper::ArrayFind(array, toFind)
{
	
	foreach(key, val in array)
	{
		if(val == toFind)
		{
			return key;
		}
	}
	return null;
}

//////////////////////////////////////////////////////////////////////
//                                                                  //
//  CLASS: ScoreList                                                //
//                                                                  //
//////////////////////////////////////////////////////////////////////
class ScoreList
{
	list = null;

	sort_max = true;
	sorted = false;

	constructor() {
		list = [];
		sort_max = true;
		sorted = false;
	}

	function Push(value, score);
	function PopMax();
	function PopMin();

	function CompareMin(a, b); // min first
	function CompareMax(a, b); // max first
}

function ScoreList::Push(value, score)
{
	list.append([value, score]);
	sorted = false;
}
function ScoreList::PopMax()
{
	if(list.len() <= 0)
	{
		return null;
	}
	if(!sorted || sort_max)
	{
		list.sort(this.CompareMin);
	}

	return list.pop()[0];
}
function ScoreList::PopMin()
{
	if(list.len() <= 0)
	{
		return null;
	}
	if(!sorted || !sort_max)
	{
		list.sort(this.CompareMax);
	}

	return list.pop()[0];
}
function ScoreList::CompareMin(a, b)
{
	if(a[1] > b[1]) 
		return 1
	else if(a[1] < b[1]) 
		return -1
	return 0;
}
function ScoreList::CompareMax(a, b)
{
	if(a[1] > b[1]) 
		return -1
	else if(a[1] < b[1]) 
		return 1
	return 0;
}

//////////////////////////////////////////////////////////////////////
//                                                                  //
//  CLASS: TownFinder                                               //
//                                                                  //
//////////////////////////////////////////////////////////////////////
class TownFinder
{
	base     = null;
	cargo    = null;
	company  = null;
	industry = null;
	map      = null;
	town     = null;
	// each item in search list is an array of these stuff:
	// [0] => town ID
	// [1] => town population
	search_list = [];

	helper = null;

	conf_min_distance = 20;
	conf_min_population = 150;

	constructor() {
		this.base = AIBase();
		this.cargo = AICargo();
		this.company = AICompany();
		this.industry = AIIndustry();
		this.map = AIMap();
		this.town = AITown();

		search_list = [];
		this.helper = Helper();
	}

	function SelectSearchList();
	function FindTwoTownsToConnectByRoad(maxDistance);
	function SelectSearchList_CompareTownByPopulation(a, b);
	function FindTwoTownsToConnectByRoad_CompareTownPairsByScore(a, b);
}

// Creates a sorted list of all towns (best first), and then shrink it to top 50.
// The top 50-list is then stored list in this.search_list.
function TownFinder::SelectSearchList(connection_list)
{
	local max_id = this.town.GetMaxTownID();
	local list = [];
	local i;
	local population;

	for(i = 0; i < max_id; ++i)
	{
		if(!this.town.IsValidTown(i) || this.helper.IsTownInConnectionList(connection_list, i))
		{
			//this.helper.Debug("Town no " + i + " is not a vaild town or allready connected, skip it");	
		}
		else
		{
			population = this.town.GetPopulation(i);
			if(population >= this.conf_min_population)
			{
				list.append([i, population]);
				//this.helper.Debug("appended town to list " + i);
			}
		}
	}
	list.sort(this.SelectSearchList_CompareTownByPopulation);
	list.reverse();

	if(list.len() > 50)
	{
		list.resize(50);
	}

	foreach(val in list)
	{
		//this.helper.Debug(val[0] + " (" + this.town.GetName(val[0]) + ") => " + val[1]);
	}

	this.search_list = list;
}

function TownFinder::SelectSearchList_CompareTownByPopulation(a, b)
{
	if(a[1] > b[1]) 
		return 1
	else if(a[1] < b[1]) 
		return -1
	return 0;
}

function TownFinder::FindTwoTownsToConnectByRoad(maxDistance)
{
	//this.helper.Debug("TownFinder::FindTwoTownsToConnectByRoad()");
	local dist_weight = 1.5;
	local pop_weight  = 1.0;

	local list_pairs = ScoreList();
	local i = 0;
	local j = 0;

	//this.helper.Debug(this.search_list.len());
	for(i = 0; i < this.search_list.len(); i++)
	{
		for(j = i+1; j < this.search_list.len(); j++)
		{
			if(i != j)
			{
				local distance = this.helper.TownDistance(this.search_list[i][0], this.search_list[j][0]);
				local total_population = this.search_list[i][1] + this.search_list[j][1];
				local score = total_population * pop_weight + 
						distance * (1 - this.helper.StepFunction(distance - maxDistance))  * dist_weight;
				if( distance > maxDistance || distance < this.conf_min_distance || total_population < this.conf_min_population )
				{
					score = 0;
				}
				else
				{
					list_pairs.Push([this.search_list[i][0], this.search_list[j][0], score, distance, total_population], score);
				}
				//this.helper.Debug("For city " + this.town.GetName(this.search_list[i][0]) + " and " + this.town.GetName(this.search_list[j][0]) + " the score is: " + score + ". The distance is: " + distance + " and the total population is: " + total_population);
			}
		}
	}

	local best_town_pair = list_pairs.PopMax();

	if(best_town_pair == null)
		return null;

	//this.helper.Debug("Best cities are: " + best_town_pair[0] + " and " + best_town_pair[1] + " with the score: " + best_town_pair[2]);
	return [ best_town_pair[0], best_town_pair[1] ];
}

function TownFinder::FindTwoTownsToConnectByRoad_CompareTownPairsByScore(a, b)
{
	if(a[2] > b[2]) 
		return 1
	else if(a[2] < b[2]) 
		return -1
	return 0;
}

//////////////////////////////////////////////////////////////////////
//                                                                  //
//  CLASS: Connection - a transport connection                      //
//                                                                  //
//////////////////////////////////////////////////////////////////////

// My aim is that this class should be as general as possible and allow 2> towns, or even cargo transport
class Connection
{
	cargo = null;

	// used towns and industries (arrays)
	town     = null;
	industry = null;

	// links to infrastructure and vehicles (arrays)
	station  = null;
	depot    = null;
	vehicle  = null;

	connection_failed = false;
	
	constructor() {
		cargo    = null;
		town     = [];
		industry = [];
		station  = [null, null];
		depot    = [null, null];
		vehicle  = [];
		connection_failed = false;
	}

	function GetTotalPoputaltion();
	function GetTotalDistance(); // only implemented for 2 towns
}

function Connection::GetTotalPoputaltion()
{
	print("GetTotalPopulation");
	local val = null;
	local pop = 0;
	foreach(val in town)
	{
		if(val != null)
		{
			pop += this.AITown().GetPopulation(val);
		}
	}
	return pop;
}
function Connection::GetTotalDistance() 
{
	print("GetTotalDistance");
	if(town.len() != 2)
		return 0;

	return this.AIMap().DistanceManhattan( this.AITown().GetLocation(town[0]), this.AITown().GetLocation(town[1]) );
}

//////////////////////////////////////////////////////////////////////
//                                                                  //
//  CLASS: CluelessAI - the AI main class                           //
//                                                                  //
//////////////////////////////////////////////////////////////////////
class CluelessAI extends AIController {
	base     = null;
	cargo    = null;
	company  = null;
	industry = null;
	map      = null;
	order    = null;
	sign     = null;
	town     = null;
	road     = null;
	vehicle  = null;
	stop     = false;

	helper   = null;
	town_finder = null;
	connection_list = [];

	state_build = false;
	state_ai_name = null;

	conf_ai_name = null
	conf_min_balance = 0;

	// All variables should be initiated with their values below!:  (the assigned values above is only there because Squirrel demands it)
	constructor() {
		this.base = AIBase();
		this.cargo = AICargo();
		this.company = AICompany();
		this.industry = AIIndustry();
		this.map = AIMap();
		this.order = AIOrder();
		this.sign = AISign();
		this.town = AITown();
		this.road = AIRoad();
		this.vehicle = AIVehicle();

		helper = Helper();
		town_finder = TownFinder();
		connection_list = [];

		state_build = false;
		state_ai_name = null;

		conf_ai_name = ["Clueless", "Cluemore", "Not so Clueless", "Quite smart", "Smart", "Intelligent", "Expert", "Geek"];
		conf_min_balance = 20000;
	}

	// WARNING: the arguments of the following functions might be wrong. Look at the implementation for the exact arguments. 
	// Squirrel don't complain if the arguments of the definitions and implementation don't match. 
	// Squirrel basically don't look at the definitions below:
	function Start();
	function Stop();

	function SetCompanyName(nameArray);

	function ManageLoan();
	function RoundLoanDown(loanAmount); // Helper
	function GetMaxMoney();

	function BuildServiceInTown();
	function BuildBusStopInTown(town);
	function BuyBuses(connection); 
	function NumBusesToBuy(connection);

	// if tile is not a road-tile it will search for closest road-tile and then start searching for a location to place it from there.
	function BuildBusStopNextToRoad(tile, min_loops, max_loops); 
	function BuildTruckStopNextToRoad(tile, min_loops, max_loops); 
	function BuildDepotNextToRoad(tile, min_loops, max_loops); 
	function BuildNextToRoad(tile, what, min_loops, max_loops);  // 'what' can be any of: "BUS_STOP", "TRUCK_STOP", "DEPOT". ('what' should be a string)

	function FindClosestRoadTile(tile, max_loops);

	function ConnectByRoad(tile1, tile2, max_loops);
	function FindRoadExtensionTile(road_tile, target_tile, min_loops, max_loops); // road_tile = start search here
	                                                                              // target_tile = when searching for a place to extend existing road, we want to get as close as possible to target_tile
																				  // min_loops = search at least for this amount of loops even if one possible extension place is found (return best found)
																				  // max_loops = maximum search loops before forced return
	function FindRoadExtensionTile_SortByDistanceToTarget(a, b); // Helper

	function ConnectByRoad(tile1, tile2);
}

function CluelessAI::Start()
{
	this.Sleep(1);

	this.state_ai_name = this.SetCompanyName(this.conf_ai_name);
	this.helper.company_name = this.state_ai_name; // Debug() need to know the (base) name of the company. 
	
	state_build = true;

	local i = 0;
	while(!this.stop)
	{
		i++;
		this.Sleep(1);

		// Sometimes...
		if(i%10)
		{
			// ... manage our loan
			ManageLoan();

			// ... check if we can afford to build some stuff
			if(this.GetMaxMoney() > 100000)
			{
				state_build = true;
			}
		}	

		if(state_build)
		{
			//this.company.SetCompanyName("[scan towns]" + this.state_ai_name);
			// Scan for two towns to connect
			local towns = null;
			local connection = Connection();
			this.town_finder.SelectSearchList(connection_list);
			towns = this.town_finder.FindTwoTownsToConnectByRoad( 100 ); // Parameter 1: max-distance

			// If we found two towns
			if(towns != null && towns.len() == 2)
			{
				//this.company.SetCompanyName("[build]" + this.state_ai_name);

				local failed = false;
				this.helper.Debug("Connect " + this.town.GetName(towns[0]) + " with " + this.town.GetName(towns[1]));

				// We don't want to worry about budgeting exactly how much money that is needed, get as much money as possible.
				// We can always pay back later. 
				this.company.SetLoanAmount(this.company.GetMaxLoanAmount());

				// Build bus-stops
				local bus_stop = [null, null];
				bus_stop[0] =  BuildBusStopInTown(towns[0]);
				this.helper.Debug("bus-stop 1 built");
				bus_stop[1] =  BuildBusStopInTown(towns[1]);
				this.helper.Debug("bus-stop 2 built");

				// Build depots
				local depot = [null, null];
				depot[0] = BuildDepotNextToRoad(bus_stop[0], 0, 100);
				this.helper.Debug("depot 1 built");
				depot[1] = BuildDepotNextToRoad(bus_stop[1], 0, 100);
				this.helper.Debug("depot 2 built");

				// Check that we built all buildings
				if(bus_stop[0] == null || bus_stop[1] == null || depot[0] == null || depot[1] == null)
				{
					failed = true;
				}

				// save town, station and depot in connection data-structure.
				connection.town.append(towns[0]);
				connection.town.append(towns[1]);
				connection.station[0] = bus_stop[0];
				connection.station[1] = bus_stop[1];
				connection.depot[0] = depot[0];
				connection.depot[1] = depot[1];
				
				local connected = false;
				this.helper.Debug("bus-stops built");
				if(!failed)
					connected = ConnectByRoad(this.road.GetRoadStationFrontTile(bus_stop[0]), this.road.GetRoadStationFrontTile(bus_stop[1]), 900) 


				// Only buy buses if we actually did connect the two cities.
				if(connected && !failed)
				{	
					this.helper.Debug("connected by road");
					// BuyBuses save the IDs of the bough buses in the connection data-structure
					BuyBuses(connection); 
					this.helper.Debug("bough buses");
				}
				else
				{
					this.helper.Debug("failed to connect by road");
					this.map.DemolishTile(connection.depot[0]);
					this.map.DemolishTile(connection.depot[1]);
					this.map.DemolishTile(connection.station[0]);
					this.map.DemolishTile(connection.station[1]);
					connection.connection_failed = true;			// store that this connection faild so we don't waste our money on buying buses for it.
				}

				// Store the connection so we don't build it again.
				connection_list.append(connection); 

				// Pay back unused money
				ManageLoan();

				state_build = false;
			}
			else
			{
				//this.helper.Debug("Could not find two towns to connect");
			}
			//this.company.SetCompanyName("" + this.state_ai_name);
		}

	}
}
function CluelessAI::Stop()
{
	this.helper.Debug("CluelessAI::Stop()");
	this.stop = true;
}

// RETURN: new name
function CluelessAI::SetCompanyName(nameArray)
{
	local i = 0;
	this.helper.Debug("i, before loop = " + i);
	while(i < nameArray.len() && !this.company.SetCompanyName(nameArray[i]))
	{
		this.helper.Debug("i++ = " + i);
		i = i + 1;
	}
	this.helper.Debug("company name done");
	return nameArray[i];
}

function CluelessAI::ManageLoan()
{
	// local constants. ( I've not found how to declare constants in Squirrel yet :( )
	local balance = this.company.GetBankBalance();
	local loan = this.company.GetLoanAmount();
	local max_balance = this.conf_min_balance + 2 * this.company.GetLoanInterval(); // Attention: max_balance is the maximum balance we accept before paying back loan.


	if( balance < this.conf_min_balance ) // bigger loan
	{
		//this.helper.Debug("ManageLoan: Want to loan more");
		//this.helper.Debug("Company: " + this.company.GetCompanyName());

		local new_loan = loan + (this.conf_min_balance - balance) + this.company.GetLoanInterval();
		if(new_loan < this.company.GetMaxLoanAmount())
			new_loan = this.company.GetMaxLoanAmount();

		new_loan = RoundLoanDown(new_loan);
		if(!this.company.SetLoanAmount(new_loan))
		{
			//this.helper.Debug(this.state_ai_name + " Failed to increase loan amount");
		}
	}
	
	else if( balance > max_balance && loan > 0) // smaller loan
	{
		//this.helper.Debug("ManageLoan: Want to pay back");
		//this.helper.Debug("Company: " + this.company.GetCompanyName());
		//this.helper.Debug("balance: " + balance);
		//this.helper.Debug("loan: " + loan);
		//this.helper.Debug("max_balance: " + max_balance);

		local pay_back = balance - max_balance;
		local new_loan = RoundLoanDown(loan - pay_back);

		//this.helper.Debug("pay_back: " + pay_back);
		//this.helper.Debug("new_loan: " + new_loan);
		
		if(!this.company.SetLoanAmount(new_loan))
		{
			//this.helper.Debug(this.state_ai_name + " Failed to decrease loan amount");
		}
			
		//this.helper.Debug("Successfully paid back loan");
	}
}

function CluelessAI::RoundLoanDown(loanAmount)
{
	return loanAmount - loanAmount % this.company.GetLoanInterval();
}
function CluelessAI::GetMaxMoney()
{
	local balance = this.company.GetBankBalance();
	local loan = this.company.GetLoanAmount();

	local max_possible_balance = balance + (this.company.GetMaxLoanAmount() - loan);

	return max_possible_balance;
}

function CluelessAI::BuildBusStopInTown(town) // this function is ugly and should ideally be removed
{
	this.helper.Debug("CluelessAI::BuildBusStopInTown(" + this.town.GetName(town) + ")");
	local location = this.town.GetLocation(town);

	return BuildBusStopNextToRoad(location, 50, 100);
}

// Build*NextToRoad functions:
// 1. Checks if 'tile' is road
// 2. If not: it locates nearby road using FindClosestRoadTile
// 3. From the starting-point on the road network, search for a suitable 
//    location to build the building. 
//    - it prefers to build the building near the starting-point
function CluelessAI::BuildBusStopNextToRoad(tile, min_loops, max_loops)
{
	return BuildNextToRoad(tile, "BUS_STOP", min_loops, max_loops);
}
function CluelessAI::BuildTruckStopNextToRoad(tile, min_loops, max_loops)
{
	return BuildNextToRoad(tile, "TRUCK_STOP", min_loops, max_loops);
}
function CluelessAI::BuildDepotNextToRoad(tile, min_loops, max_loops)
{
	return BuildNextToRoad(tile, "DEPOT", min_loops, max_loops);
}
function CluelessAI::BuildNextToRoad(tile, what, min_loops, max_loops)
{
	local start_tile = tile;

	local green_list = ScoreList();
	local red_list = [];

	local found_locations = ScoreList();

	local i = 0;
	local ix, iy;
	local curr_x, curr_y;
	local curr_tile;
	local adjacent_x, adjacent_y;
	local adjacent_tile;

	local adjacent_loop = [ [-1,0], [1,0], [0,-1], [0,1] ];

	if(!start_tile)
		return false;

	if(!this.road.IsRoadTile(start_tile))
	{
		start_tile = FindClosestRoadTile(start_tile, 4);
		if(!start_tile)
		{
			return false;
		}
	}

	curr_tile = start_tile;

	while( i++ < 100 )
	{
		curr_x = this.map.GetTileX(curr_tile);
		curr_y = this.map.GetTileY(curr_tile);

		
		// scan adjacent tiles

		foreach(adjacent_offset in adjacent_loop)
		{
			adjacent_x = curr_x + adjacent_offset[0];
			adjacent_y = curr_y + adjacent_offset[1];

			adjacent_tile = this.map.GetTileIndex(adjacent_x, adjacent_y);

			if(this.road.AreRoadTilesConnected(curr_tile, adjacent_tile))
			{
				if( this.helper.ArrayFind(red_list, adjacent_tile) == null )
				{
					green_list.Push(adjacent_tile, this.map.DistanceSquare(adjacent_tile, start_tile));
					//this.sign.BuildSign(adjacent_tile, i+":"+this.map.DistanceSquare(adjacent_tile, start_tile));
					this.Sleep(1);
				}
			}
			else if(this.road.BuildRoad(adjacent_tile, curr_tile))
			{
				//this.sign.BuildSign(adjacent_tile, i+"|"+this.map.DistanceSquare(adjacent_tile, start_tile));

				local ret;

				if(what == "BUS_STOP")
				{
					local truck = false; 
					local drive_tough = false;
					ret = this.road.BuildRoadStation(adjacent_tile, curr_tile, truck, drive_tough)
				}
				else if(what == "TRUCK_STOP")
				{
					local truck = true; 
					local drive_tough = false;
					ret = this.road.BuildRoadStation(adjacent_tile, curr_tile, truck, drive_tough)
				}
				else if(what == "DEPOT")
				{
					ret = this.road.BuildRoadDepot(adjacent_tile, curr_tile)
				}
				else
				{
					this.helper.Debug("ERROR: Invalid value of argument 'what' to function BuildNextToRoad(tile, what)");
					this.road.RemoveRoad(adjacent_tile, curr_tile);
					return null;
				}

				if(ret)
				{
					found_locations.Push(adjacent_tile, this.map.DistanceSquare(adjacent_tile, start_tile));
				}
				else
				{
					this.road.RemoveRoad(adjacent_tile, curr_tile);
				}
			}
			else
			{
				//this.sign.BuildSign(adjacent_tile, "x");
			}

			this.Sleep(1);
		}

		// if found at least one location and we have looped at least min_loops. => don't search more
		if(found_locations.list.len() > 0 && i >= min_loops)
		{
			break;
		}

		red_list.append(curr_tile);

		// select best tile from green_list
		curr_tile = green_list.PopMin();

		if(!curr_tile || !this.map.IsValidTile(curr_tile)) // if green_list was empty
		{
			break;
		}
	}

	// get best built building
	local building_tile = found_locations.PopMin();
	if(building_tile == null) // return null, if no location at all was found.
	{
		this.helper.Debug("BuildNextToRoad: failed to build: " + what);
		return null;
	}

	// remove the other buildings + the road-bit that connect them to the network
	// the later is good to remove the ugly stumps that else would be left and irritate human players. 
	// There is however a case when the stumps can't be removed and that if a building is connected to another company's road.
	local val = null;
	local front = null;
	while( (val = found_locations.PopMin()) != null )
	{
		this.Sleep(1);
		if(what == "BUS_STOP" || what == "TRUCK_STOP")
		{
			front = this.road.GetRoadStationFrontTile(val);	
			this.map.DemolishTile(val);
			this.road.RemoveRoad(front, val)
		}
		else if(what == "DEPOT")
		{
			front = this.road.GetRoadDepotFrontTile(val);	
			this.map.DemolishTile(val);
			this.road.RemoveRoad(front, val)
		}
		else
		{
			this.map.DemolishTile(val);
		}
	}

	this.Sleep(10);
	helper.ClearAllSigns();
	
	// return the location of the best and only remaining building.
	return building_tile;
	
}

function CluelessAI::FindClosestRoadTile(tile, max_radius)
{
	if(!tile || !this.map.IsValidTile(tile))
		return null;

	if(this.road.IsRoadTile(tile))
		return tile;
	
	local r; // current radius

	local start_x = this.map.GetTileX(tile);
	local start_y = this.map.GetTileY(tile);

	local x0, x1, y0, y1;
	local ix, iy;
	local test_tile;

	for(r = 1; r < max_radius; ++r)
	{
		y0 = start_y - r;
		y1 = start_y + r;
		for(ix = start_x - r; ix <= start_x + r; ++ix)
		{
			test_tile = this.map.GetTileIndex(ix, y0)
			if(test_tile != null && this.road.IsRoadTile(test_tile))
				return test_tile;

			test_tile = this.map.GetTileIndex(ix, y1)
			if(test_tile != null && this.road.IsRoadTile(test_tile))
				return test_tile;
		}

		x0 = start_x - r;
		x1 = start_x + r;
		for(iy = start_y - r + 1; iy <= start_y + r - 1; ++iy)
		{
			test_tile = this.map.GetTileIndex(x0, iy)
			if(test_tile != null && this.road.IsRoadTile(test_tile))
				return test_tile;

			test_tile = this.map.GetTileIndex(x1, iy)
			if(test_tile != null && this.road.IsRoadTile(test_tile))
				return test_tile;

		}
	}

	return null;
}

function CluelessAI::FindRoadExtensionTile(road_tile, target_tile, min_loops, max_loops)
{
	// road_tile belong to a road-network of one or more tiles.
	// This function aims to find a tile in this network from which the network can be extended.
	// It stops the search when either:
	// * the whole network have been scanned
	// * it have looped for min_loops times AND have found at least one solution
	// * it have looped for max_loops times
	//
	// Returns the best solution-tile.


	if(road_tile == null || !this.road.IsRoadTile(road_tile) || target_tile == null || !this.map.IsValidTile(target_tile))
		return null;


	local start = road_tile;

	local green_list = ScoreList()
	local red_list = [];

	// TODO: Use ScoreList instead of a plain array.
	local extend_list = []; // each item is an array: [tile, MH distance to target]

	local i = 0;
	local ix, iy;
	local curr_x, curr_y;
	local curr_tile;
	local adjacent_x, adjacent_y;
	local adjacent_tile;

	local adjacent_loop = [ [-1,0], [1,0], [0,-1], [0,1] ];

	curr_tile = start;

	while( i++ < max_loops )
	{
		curr_x = this.map.GetTileX(curr_tile);
		curr_y = this.map.GetTileY(curr_tile);

		//this.sign.BuildSign(curr_tile, "i:" + i)
		
		// scan adjacent tiles

		foreach(adjacent_offset in adjacent_loop)
		{
			adjacent_x = curr_x + adjacent_offset[0];
			adjacent_y = curr_y + adjacent_offset[1];

			adjacent_tile = this.map.GetTileIndex(adjacent_x, adjacent_y);

			if(this.road.AreRoadTilesConnected(curr_tile, adjacent_tile))
			{
				// special case: we have reached the target-tile.
				if(adjacent_tile == target_tile)
				{
					return adjacent_tile;
				}

				// add tile to green_list if it is not in red_list
				if( this.helper.ArrayFind(red_list, adjacent_tile) == null )
				{
					green_list.Push(adjacent_tile, this.map.DistanceManhattan(adjacent_tile, target_tile));
				}
			}
			else 
			{
				//this.helper.Debug("CluelessAI::FindRoadExtensionTile: found one solution");
				// Perhaps the road can be extended here. 
				// Test if it is possible to do so.
				local test_mode = AITestMode();
				if(this.road.BuildRoad(curr_tile, adjacent_tile))
				{
					//this.helper.Debug("Possible to build road");
					extend_list.append( [ curr_tile, this.map.DistanceManhattan(curr_tile, target_tile) ] );
				}
			}

			this.Sleep(1);
		}

		// stop loop if at least one solution found and i > min_loops
		if(extend_list.len() > 0 && i > min_loops)
		{
			break;
		}

		red_list.append(curr_tile);

		// select best tile from green_list
		curr_tile = green_list.PopMin();

		if(!curr_tile || !this.map.IsValidTile(curr_tile)) // if green_list was empty
		{
			break;
		}
	}

	if(extend_list.len() == 0)
	{
		this.helper.Debug("CluelessAI::FindRoadExtensionTile: found zero solutions");
		return null;
	}
	else if(extend_list.len() == 1)
		return extend_list[0][0];
	else
	{
		extend_list.sort(FindRoadExtensionTile_SortByDistanceToTarget);
		return extend_list[0][0];
	}

	return null;
}
function CluelessAI::FindRoadExtensionTile_SortByDistanceToTarget(a, b)
{
	if(a[1] > b[1]) 
		return 1
	else if(a[1] < b[1]) 
		return -1
	return 0;
}

function CluelessAI::ConnectByRoad(tile1, tile2, max_loops)
{
	if(tile1 == null || tile2 == null || max_loops == null)
		return false;

	if(!this.road.IsRoadTile(tile1) || !this.road.IsRoadTile(tile2))
		return false;
	
	local start_tile = [null, null];

	/*this.sign.BuildSign(tile1, "tile1");
	this.sign.BuildSign(tile2, "tile2");*/

	// Assuming that tile1, tile2 might be in the middle of a city, we make
	// some scanning for good road-tiles in the 'border' of the road network 
	// that tile1, tile2 might be connected to.
	//
	// Update: Now we actually trying to find a way to go from town 0 to town 1
	//         using existing road. If that fails we get an extension-point of
	//         the second town too and continue with building road.
	start_tile[0] = FindRoadExtensionTile(tile1, tile2, 300, 400);  // scan from,  scan to,  min loops,  max loops
	if(start_tile[0] == tile2)
		return true; // return true if the other city was found.

	start_tile[1] = FindRoadExtensionTile(tile2, tile1, 40, 100);
	if(start_tile[1] == tile1)
		return true; // return true if the other city was found.


	// Check that the two start-tiles are valid.
	if(!start_tile[0] || !this.map.IsValidTile(start_tile[0]) || !start_tile[1] || !this.map.IsValidTile(start_tile[1]))
		return false;
	

	// Now we have two tiles, start_tile[0] and start_tile[1] to connect by road, and both have 
	// adjacent tiles which we can extend into. (The FindRoadExtensionTile-function make sure 
	// that there are adjacent tiles that we can extend to.)

	// The basic idea is to build from curr_tile[0] to curr_tile[1]. There is no green/red-list.
	// It will just select a direction to head at, and then continue until it hits something. 
	// Then it will try the other direction that heads towards curr_tile[1]. If that also fails,
	// it have some minor hacks for minor obstacles. Bigger things like industries can make the 
	// algorithm fail. 
	
	local curr_tile =  start_tile;
	local i = 0;
	local dirr_x = 1;
	local dirr_y = 0;
	local go_x = true;
	local diff_x = null;
	local diff_y = null;
	local target_x = null;
	local target_y = null;
	local next_tile = null;
	local built_road = false;
	local fails = 0;

	diff_x = this.map.GetTileX(curr_tile[1]) - this.map.GetTileX(curr_tile[0]);
	diff_y = this.map.GetTileY(curr_tile[1]) - this.map.GetTileY(curr_tile[0]);

	if(diff_x > 0)
		dirr_x = 1;
	else
		dirr_x = -1;

	if(diff_y > 0)
		dirr_y = 1;
	else
		dirr_y = -1;

	target_x = this.map.GetTileX(curr_tile[1]);
	target_y = this.map.GetTileY(curr_tile[1]);
	if(abs(diff_x) > abs(diff_y))
		go_x = true;
	else
		go_x = false;


	while(curr_tile[0] != curr_tile[1] && ++i <= max_loops)
	{
		if(go_x)
		{
			//this.helper.Debug("ConnectByRoad: go_x");
			next_tile = this.map.GetTileIndex(this.map.GetTileX(curr_tile[0]) + dirr_x, this.map.GetTileY(curr_tile[0]));

			built_road = this.road.BuildRoad(curr_tile[0], next_tile);

			//if(built_road)
			if(built_road || (fails < 1 && this.road.AreRoadTilesConnected(curr_tile[0], next_tile)) )
			{
				/*if(!built_road)
					this.sign.BuildSign(next_tile, "jump");
				else
					this.sign.BuildSign(curr_tile[0], "ok");*/
				curr_tile[0] = next_tile;
				fails = 0;
			}
			else
			{
				//this.sign.BuildSign(curr_tile[0], "x");
				fails++;
				if(fails == 1)
					go_x = false;
				else
				{
					// try build one step in wrong direction
					next_tile = this.map.GetTileIndex(this.map.GetTileX(curr_tile[0]) - dirr_x, this.map.GetTileY(curr_tile[0]));
					if(this.road.BuildRoad(curr_tile[0], next_tile))
					{
						curr_tile[0] = next_tile;
						go_x = false;

						// don't reset 'fails', a non-fail is building in the right direction  (if you don't understand this comment, read the previous one too!)
					}
					else
						break; // bail out
				}
			}
			if((dirr_x == 1 && this.map.GetTileX(next_tile) >= target_x) || (dirr_x == -1 && this.map.GetTileX(next_tile) <= target_x))
			{
				go_x = false;
			}
		}
		else if(!go_x)  // must be else, otherwise we might miss the target!
		{
			//this.helper.Debug("ConnectByRoad: !go_x");
			next_tile = this.map.GetTileIndex(this.map.GetTileX(curr_tile[0]), this.map.GetTileY(curr_tile[0]) + dirr_y);

			built_road = this.road.BuildRoad(curr_tile[0], next_tile);

			//if(built_road)
			if(built_road || (fails < 1 && this.road.AreRoadTilesConnected(curr_tile[0], next_tile)) )
			{
				/*if(!built_road)
					this.sign.BuildSign(next_tile, "jump");
				else
					this.sign.BuildSign(curr_tile[0], "ok");*/
				curr_tile[0] = next_tile;
				fails = 0;
			}
			else
			{
				//this.sign.BuildSign(curr_tile[0], "x");
				fails++;
				if(fails == 1)
					go_x = true;
				else
				{
					// try build one step in wrong direction
					next_tile = this.map.GetTileIndex(this.map.GetTileX(curr_tile[0]), this.map.GetTileY(curr_tile[0]) - dirr_y);
					if(this.road.BuildRoad(curr_tile[0], next_tile))
					{
						curr_tile[0] = next_tile;
						go_x = true;
						// don't reset fails, a non-fail is building in the right direction
					}
					else
						break; // bail out
				}
			}
			if((dirr_y == 1 && this.map.GetTileY(next_tile) >= target_y) || (dirr_y == -1 && this.map.GetTileY(next_tile) <= target_y))
			{
				go_x = true;
			}
		}

		// Update dirr_*. I'm not sure if it is good to do so, the rest of the code (above) was written as if dirr_* are constant.
		// begin
		diff_x = this.map.GetTileX(curr_tile[1]) - this.map.GetTileX(curr_tile[0]);
		diff_y = this.map.GetTileY(curr_tile[1]) - this.map.GetTileY(curr_tile[0]);

		if(diff_x > 0)
			dirr_x = 1;
		else
			dirr_x = -1;

		if(diff_y > 0)
			dirr_y = 1;
		else
			dirr_y = -1;
		// end
	}

	if(curr_tile[0] == curr_tile[1])
		return true;

	return false;
}

function CluelessAI::BuyBuses(connection)
{
	if(connection.station.len() != 2)
	{
		return null;
	}
	this.helper.Debug("BuyBuses");
	local cargo = 0; // passengers
	local bus_model = null;
	local min_reliability = 85;
	local new_bus;
	local i;
	local n_buses = 2
	
	this.helper.Debug("bus");
	n_buses = NumBusesToBuy(connection);
	this.helper.Debug("bus++");

	// find a bus model to buy
	while(bus_model == null || !this.vehicle.IsValidEngine(bus_model) && min_reliability > 0)
	{
		bus_model = this.vehicle.FindBestRoadVehicle(cargo, min_reliability);
		min_reliability -= 10;
	}
	//bus_model = this.vehicle.FindBestRoadVehicle(cargo, 0);
	if(bus_model == null || !this.vehicle.IsValidEngine(bus_model))
	{
		this.helper.Debug("BuyBuses: failed to find a bus to build");
		return false;
	}
	
	this.helper.Debug("bus2");

	// Buy buses
	for(i = 0; i < n_buses; ++i)
	{
		new_bus = this.vehicle.BuildVehicle(connection.depot[i%2], bus_model);
		if(new_bus != null && this.vehicle.IsValidVehicle(new_bus))
		{
			this.helper.Debug("I built a bus");
		}

		if(i>0)
		{
			this.order.ShareOrders(new_bus, connection.vehicle[0]);
		}
		connection.vehicle.append(new_bus);
	}

	this.helper.Debug("bus3");
	
	if(connection.vehicle[0] == null || !this.vehicle.IsValidVehicle(connection.vehicle[0]))
		this.helper.Debug("Vehicle[0] is not valid!");

	this.order.AppendOrder(connection.vehicle[0], connection.depot[0], this.order.AIOF_SERVICE_IF_NEEDED);
	this.order.AppendOrder(connection.vehicle[0], connection.station[0], this.order.AIOF_NONE); 
	this.order.AppendOrder(connection.vehicle[0], connection.depot[0], this.order.AIOF_SERVICE_IF_NEEDED);  
	this.order.AppendOrder(connection.vehicle[0], connection.station[1], this.order.AIOF_NONE);
	
	for(i = 0; i < n_buses; ++i)
	{
		if(i%2 == 1)
		{
			this.vehicle.SkipVehicleOrder(connection.vehicle[i]);
			this.Sleep(350);
		}
		this.vehicle.StartStopVehicle(connection.vehicle[i]);

	}

	return connection;
}

function CluelessAI::NumBusesToBuy(connection)
{
	this.helper.Debug("NumBusesToBuy()");
	local population = connection.GetTotalPoputaltion().tofloat();
	local distance = connection.GetTotalDistance().tofloat();

	this.helper.Debug("NumBusesToBuy(): distance between towns: " + distance);
	this.helper.Debug("NumBusesToBuy(): total town population:  " + population);

	local num_bus = (population / 400.0 );
	local extra = distance/ 80.0;
	num_bus += extra;

	this.helper.Debug("Buy " + num_bus.tointeger() + " buses");
	return num_bus.tointeger();
}


//////////////////////////////////////////////////////////////////////
//                                                                  //
//  CLASS: FCluelessAI - the factory class - registration of the ai //
//                                                                  //
//////////////////////////////////////////////////////////////////////
class FCluelessAI extends AIFactory {
	function GetAuthor()      { return "Zuu"; }
	function GetName()        { return "Clueless"; }
	function GetDescription() { return "An AI that connects town-pairs with bus-links"; }
	function GetVersion()     { return 1; }
	function GetDate()        { return "2007-04-11"; }
	function CreateInstance() { return "CluelessAI"; }
}

/* Tell the core we are an AI */
iFCluelessAI <- FCluelessAI();
